// To compile this code:
//    Move it to the main Electric directory
//    Run "Command prompt"
//    cd to the main Electric directory
//    type "javac -classpath lib\java test.java"
//
// then, inside of Electric's Java interpreter, type:
//    test.ex1()
//    test.ex2()
//    test.ex3()
//    test.ex4()
//    test.doVoid()/Boolean/Byte/Char/Short/Int/Long/Float/Double/String

import COM.staticfreesoft.*;


public class test
{
	final static public int STATEBITS = 036000000000;
	final static public int INPORT    = 020000000000;
	final static public int OUTPORT   = 022000000000;
	final static public int BIDIRPORT = 024000000000;
	final static public int PWRPORT   = 026000000000;
	final static public int GNDPORT   = 030000000000;

	public static Object ex1()
	{
		/* create a cell called "tran-contact" in the current library */
		Electric.NodeProto mycell = Electric.newNodeProto("tran-contact", Electric.curLib());

        /* get pointers to primitives */
		Electric.NodeProto tran = Electric.getNodeProto("P-Transistor");
		Electric.NodeProto contact = Electric.getNodeProto("Metal-1-Polysilicon-1-Con");

        /* get default sizes of these primitives */
		int tlowx = ((Integer)Electric.getVal(tran, "lowx")).intValue();
		int thighx = ((Integer)Electric.getVal(tran, "highx")).intValue();
		int tlowy = ((Integer)Electric.getVal(tran, "lowy")).intValue();
		int thighy = ((Integer)Electric.getVal(tran, "highy")).intValue();
		int clowx = ((Integer)Electric.getVal(contact, "lowx")).intValue();
		int chighx = ((Integer)Electric.getVal(contact, "highx")).intValue();
		int clowy = ((Integer)Electric.getVal(contact, "lowy")).intValue();
		int chighy = ((Integer)Electric.getVal(contact, "highy")).intValue();

        /* get pointer to Polysilicon arc and its default width */
		Electric.ArcProto arctype = Electric.getArcProto("Polysilicon-1");
		int width = ((Integer)Electric.getVal(arctype, "nominalwidth")).intValue();

        /* create the transistor and the contact to its left */
		Electric.NodeInst c1 = Electric.newNodeInst(contact, clowx, chighx, clowy, chighy,
			0, 0, mycell);
		Electric.NodeInst t1 = Electric.newNodeInst(tran, tlowx+8000, thighx+8000,
			tlowy, thighy, 0, 0, mycell);

		/* get the transistor's left port coordinates */
		Electric.PortProto tport = Electric.getPortProto(tran, "p-trans-poly-left");
		Integer[] tpos = Electric.portPosition(t1, tport);

		/* get the contacts's only port coordinates */
		Electric.PortProto cport = (Electric.PortProto)Electric.getVal(contact, "firstportproto");
		Integer[] cpos = Electric.portPosition(c1, cport);

		/* run a wire between the primitives */
		Electric.ArcInst ai = Electric.newArcInst(arctype, width, 0,
			t1, tport, tpos[0].intValue(), tpos[1].intValue(),
			c1, cport, cpos[0].intValue(), cpos[1].intValue(), mycell);

		/* make ports on the transistor */
		Electric.PortProto topport = Electric.getPortProto(tran, "p-trans-diff-top");
		Electric.PortProto botport = Electric.getPortProto(tran, "p-trans-diff-bottom");
		Electric.PortProto p1 = Electric.newPortProto(mycell, t1, topport, "topdiff");
		Electric.PortProto p2 = Electric.newPortProto(mycell, t1, botport, "botdiff");

		int userbits = ((Integer)Electric.getVal(p1, "userbits")).intValue();
		Electric.setVal(p1, "userbits", new Integer((userbits & ~STATEBITS) | INPORT), 0);
		userbits = ((Integer)Electric.getVal(p2, "userbits")).intValue();
		Electric.setVal(p2, "userbits", new Integer((userbits & ~STATEBITS) | OUTPORT), 0);

		/* name the arc */
		Electric.setVal(ai, "ARC_name", "Connection", Electric.vdisplay);
 
        /* create a pure-layer node with outline information */
		Electric.NodeProto m1node = Electric.getNodeProto("Metal-1-Node");
		Electric.NodeInst mn = Electric.newNodeInst(m1node, 3000, 5000, 2000, 4000,
			0, 0, mycell);
		int[] outline = new int[8];
		outline[0] = -1000;   outline[1] = 0;
		outline[2] = 0;       outline[3] = 1000;
		outline[4] = 1000;    outline[5] = 0;
		outline[6] = 0;       outline[7] = -1000;
		Electric.setVal(mn, "trace", outline, 0);

		System.out.println("Created the cell 'tran-contact'");
		return(mycell);
	}

	public static Object ex2()
	{
		/* create a cell called "two-trans" */
		Electric.NodeProto highercell = Electric.newNodeProto("two-trans", Electric.curLib());

		/* get pointer to the "tran-contact" cell */
		Electric.NodeProto tc = Electric.getNodeProto("tran-contact");

		/* get size of this cell */
		int lowx = ((Integer)Electric.getVal(tc, "lowx")).intValue();
		int highx = ((Integer)Electric.getVal(tc, "highx")).intValue();
		int lowy = ((Integer)Electric.getVal(tc, "lowy")).intValue();
		int highy = ((Integer)Electric.getVal(tc, "highy")).intValue();

		/* create the two cell instances, one above the other */
		Electric.NodeInst o1 = Electric.newNodeInst(tc, lowx, highx, lowy, highy,
			0, 0, highercell);
		Electric.NodeInst o2 = Electric.newNodeInst(tc, lowx, highx, lowy+10000, highy+10000,
			0, 0, highercell);

		/* get pointer to P-Active arc and its default width */
		Electric.ArcProto darctype = Electric.getArcProto("P-Active");
		int dwidth = ((Integer)Electric.getVal(darctype, "nominalwidth")).intValue();

		/* get the bottom cell's top port */
		Electric.PortProto lowport = Electric.getPortProto(tc, "topdiff");
		Integer[] lowpos = Electric.portPosition(o1, lowport);

		/* get the top cell's bottom port */
		Electric.PortProto highport = Electric.getPortProto(tc, "botdiff");
		Integer[] highpos = Electric.portPosition(o2, highport);

		/* run a wire between the primitives */
		Electric.ArcInst ai = Electric.newArcInst(darctype, dwidth, 0,
			o1, lowport, lowpos[0].intValue(), lowpos[1].intValue(),
			o2, highport, highpos[0].intValue(), highpos[1].intValue(), highercell);

		System.out.println("Created the cell 'two-trans'");

		/* invoke the "help" command */
		Electric.Tool user = Electric.getTool("user");
		String[] message = new String[2];
		message[0] = "show";
		message[1] = "technologies";
		Electric.tellTool(user, 2, message);
		return(highercell);
	}

	public static String describenode(Electric.NodeInst node)
	{
		Electric.NodeProto proto = (Electric.NodeProto)Electric.getVal(node, "proto");
		int primindex = ((Integer)Electric.getVal(proto, "primindex")).intValue();
		if (primindex == 0)
		{
			return( (String) Electric.getVal( (Electric.Cell)Electric.getVal(proto, "cell"), "cellname"));
		}
		return((String)Electric.getVal(proto, "primname"));
	}

	public static void ex3()
	{
		Electric.NodeInst ni;
		Electric.NodeProto mycell = Electric.getNodeProto("tran-contact");

		for(ni = (Electric.NodeInst)Electric.getVal(mycell, "firstnodeinst");
			!ni.isNull();
			ni = (Electric.NodeInst)Electric.getVal(ni, "nextnodeinst"))
		{
			System.out.println("Found node " + describenode(ni));
		}
	}

	public static void ex4()
	{
		Electric.NodeProto mycell = Electric.getNodeProto("tran-contact");
		int key = Electric.initSearch(2000, 10000, -3000, 3000, mycell);
		for(;;)
		{
			Electric.Geom object = (Electric.Geom)Electric.nextObject(key);
			if (object.isNull()) break;
			int isnode = ((Integer)Electric.getVal(object, "entryisnode")).intValue();
			if (isnode != 0)
			{
				Electric.NodeInst ni = (Electric.NodeInst)Electric.getVal(object, "entryaddr");
				System.out.println("Found node " + describenode(ni));
			} else
			{
				Electric.ArcInst ai = (Electric.ArcInst)Electric.getVal(object, "entryaddr");
				String arcname = (String)Electric.getVal((Electric.ArcProto)Electric.getVal(ai, "proto"), "protoname");
				System.out.println("Found arc " + arcname);
			}
		}
	}

	public static void    doVoid() {}
	public static boolean doBoolean() { return(true); }
	public static byte    doByte() { return(24); }
	public static char    doChar() { return('A'); }
	public static short   doShort() { return(5000); }
	public static int     doInt() { return(2000000); }
	public static long    doLong() { return(123456789); }
	public static float   doFloat() { return(3.14159f); }
	public static double  doDouble() { return(3.1415926535); }
	public static String  doString() { return("Howdy"); }
	public static void    doExit() { System.exit(12); }
}
