/* -*- tab-width: 4 -*- */

/* Alloca() simulation.
 */

#include "elkkernel.h"

#ifndef USE_ALLOCA

/* extern char *malloc(); */

MEM_NODE *Mem_List;

CHAR *Mem_Alloc (unsigned int size) {
	CHAR *ret;

	Disable_Interrupts;
	if ((ret = (CHAR *)malloc (size)) == 0)
		Fatal_Error (x_("not enough memory to malloc %u bytes"), size);
	Enable_Interrupts;
	return ret;
}

void Free_Mem_Nodes (MEM_NODE *first) {
	MEM_NODE *p;

	Disable_Interrupts;
	while ((p = first) != 0) {
		first = first->next;
		if (--p->refcnt == 0)
			free ((CHAR *)p);
	}
	Enable_Interrupts;
}

void Save_Mem_Nodes (ELKObject cont) {
	unsigned sum = 0;
	CHAR *s;
	MEM_NODE *p;
	ELKObject str;
	GC_Node;

	CONTROL(cont)->memlist = Mem_List;
	for (p = Mem_List; p; p = p->next)
		sum += p->len;
	GC_Link (cont);
	str = Make_String ((CHAR *)0, sum);
	CONTROL(cont)->memsave = str;
	GC_Unlink;
	for (p = Mem_List, s = STRING(str)->data; p; s += p->len, p = p->next) {
		bcopy ((CHAR *)(p+1), s, p->len);
		p->refcnt++;
	}
}

void Restore_Mem_Nodes (ELKObject cont) {
	MEM_NODE *p;
	CHAR *s;
	ELKObject str;

	Free_Mem_Nodes (Mem_List);
	Mem_List = CONTROL(cont)->memlist;
	str = CONTROL(cont)->memsave;
	for (p = Mem_List, s = STRING(str)->data; p; s += p->len, p = p->next) {
		p->refcnt++;
		bcopy (s, (CHAR *)(p+1), p->len);
	}
}

ELKObject Save_GC_Nodes (void) {
	ELKObject vec;
	register unsigned sum = 0, i = 0, n;
	register GCNODE *p;

	for (p = GC_List; p; p = p->next)
		sum += p->gclen <= 0 ? 1 : p->gclen-1;
	vec = Make_Vector (sum, Null);
	for (p = GC_List; p; p = p->next, i += n) {
		n = p->gclen <= 0 ? 1 : p->gclen-1;
		bcopy ((CHAR *)p->gcobj, (CHAR *)&(VECTOR(vec)->data[i]),
			n * sizeof (ELKObject));
	}
	return vec;
}

void Restore_GC_Nodes (ELKObject vec) {
	register int i = 0, n;
	register GCNODE *p;

	for (p = GC_List; p; p = p->next, i += n) {
		n = p->gclen <= 0 ? 1 : p->gclen-1;
		bcopy ((CHAR *)&(VECTOR(vec)->data[i]), (CHAR *)p->gcobj,
			n * sizeof (ELKObject));
	}
}

#endif
