/* -*- tab-width: 4 -*- */

#include <ctype.h>

#include "elkkernel.h"

ELKObject Obarray;
int Hash (CHAR *str, int len);
ELKObject Make_Symbol (ELKObject name);
ELKObject Obarray_Lookup (CHAR *str, int len);

ELKObject Null,
	   Trueob,
	   Falseob,
	   False2,
	   Unbound,
	   Special,
	   Void,
	   Newline,
	   Eof,
	   Zero,
	   One;

void Init_Symbol (void) {
	SET(Null, T_Null, 0);
	SET(Trueob, T_Boolean, 1);
	SET(Falseob, T_Boolean, 0);
	False2 = Falseob;
	SET(Unbound, T_Unbound, 0);
	SET(Special, T_Special, 0);
	SET(Eof, T_End_Of_File, 0);
	Newline = Make_Char ('\n');
	Zero = Make_Integer (0);
	One = Make_Integer (1);
	Obarray = Make_Vector (OBARRAY_SIZE, Null);
	Global_GC_Link (Obarray);
	Define_Symbol (&Void, x_(""));
}

ELKObject Make_Symbol (ELKObject name) {
	ELKObject sym;
	register struct S_Symbol *sp;
	GC_Node;

	GC_Link (name);
	sym = Alloc_Object (sizeof (struct S_Symbol), T_Symbol, 0);
	sp = SYMBOL(sym);
	sp->name = name;
	sp->value = Unbound;
	sp->plist = Null;
	GC_Unlink;
	return sym;
}

ELKObject P_Symbolp (ELKObject x) {
	return TYPE(x) == T_Symbol ? Trueob : Falseob;
}

ELKObject P_Symbol_To_String (ELKObject x) {
	Check_Type (x, T_Symbol);
	return SYMBOL(x)->name;
}

ELKObject Obarray_Lookup (register CHAR *str, register int len) {
	register int h;
	register struct S_String *s;
	register struct S_Symbol *sym;
	ELKObject p;

	h = Hash (str, len) % OBARRAY_SIZE;
	for (p = VECTOR(Obarray)->data[h]; !Nullp (p); p = sym->next) {
		sym = SYMBOL(p);
		s = STRING(sym->name);
		if (s->size == len && bcmp (s->data, str, len) == 0)
			return p;
	}
	return Make_Integer (h);
}

ELKObject CI_Intern (const CHAR *str) {
	ELKObject s, *p, sym, ostr;
	register int len;
	register const CHAR *src;
	CHAR *dst;
	CHAR buf[128];
	Alloca_Begin;

	len = estrlen (str);
	if (len > sizeof (buf)) {
		Alloca (dst, CHAR*, len);
	} else
		dst = buf;
	src = str;
	str = dst;
	for ( ; *src; src++, dst++)
		*dst = isupper (*src) ? tolower (*src) : *src;
	s = Obarray_Lookup ((CHAR *)str, len);
	if (TYPE(s) != T_Fixnum) {
		Alloca_End;
		return s;
	}
	ostr = Make_Const_String (str, len);
	sym = Make_Symbol (ostr);
	p = &VECTOR(Obarray)->data[FIXNUM(s)];
	SYMBOL(sym)->next = *p;
	Alloca_End;
	*p = sym;
	return sym;
}

ELKObject Intern (const CHAR *str) {
	ELKObject s, *p, sym, ostr;
	register int len;

	if (Case_Insensitive)
		return CI_Intern (str);
	len = estrlen (str);
	s = Obarray_Lookup ((CHAR *)str, len);
	if (TYPE(s) != T_Fixnum)
		return s;
	ostr = Make_Const_String (str, len);
	sym = Make_Symbol (ostr);
	p = &VECTOR(Obarray)->data[FIXNUM(s)];
	SYMBOL(sym)->next = *p;
	*p = sym;
	return sym;
}

ELKObject P_String_To_Symbol (ELKObject str) {
	ELKObject s, *p, sym;

	Check_Type (str, T_String);
	s = Obarray_Lookup (STRING(str)->data, STRING(str)->size);
	if (TYPE(s) != T_Fixnum)
		return s;
	str = Make_String (STRING(str)->data, STRING(str)->size);
	sym = Make_Symbol (str);
	p = &VECTOR(Obarray)->data[FIXNUM(s)];
	SYMBOL(sym)->next = *p;
	*p = sym;
	return sym;
}

ELKObject P_Oblist (void) {
	register int i;
	ELKObject p, list, bucket;
	GC_Node2;

	p = list = Null;
	GC_Link2 (p, list);
	for (i = 0; i < OBARRAY_SIZE; i++) {
		bucket = Null;
		for (p = VECTOR(Obarray)->data[i]; !Nullp (p); p = SYMBOL(p)->next)
			bucket = Cons (p, bucket);
		if (!Nullp (bucket))
			list = Cons (bucket, list);
	}
	GC_Unlink;
	return list;
}

ELKObject P_Put (int argc, ELKObject *argv) {
	ELKObject sym, key, last, tail, prop;
	GC_Node3;

	sym = argv[0];
	key = argv[1];
	Check_Type (sym, T_Symbol);
	Check_Type (key, T_Symbol);
	last = Null;
	for (tail = SYMBOL(sym)->plist; !Nullp (tail); tail = Cdr (tail)) {
		prop = Car (tail);
		if (EQ(Car (prop), key)) {
			if (argc == 3)
				Cdr (prop) = argv[2];
			else if (Nullp (last))
				SYMBOL(sym)->plist = Cdr (tail);
			else
				Cdr (last) = Cdr (tail);
			return key;
		}
		last = tail;
	}
	if (argc == 2)
		return Falseob;
	GC_Link3 (sym, last, key);
	tail = Cons (key, argv[2]);
	tail = Cons (tail, Null);
	if (Nullp (last))
		SYMBOL(sym)->plist = tail;
	else
		Cdr (last) = tail;
	GC_Unlink;
	return key;
}

ELKObject P_Get (ELKObject sym, ELKObject key) {
	ELKObject prop;

	Check_Type (sym, T_Symbol);
	Check_Type (key, T_Symbol);
	prop = Assq (key, SYMBOL(sym)->plist);
	if (!Truep (prop))
		return Falseob;
		/*
		 * Do we want to signal an error or return #f?
		 *
		 * Primitive_Error ("~s has no such property: ~s", sym, key);
		 */
	return Cdr (prop);
}

ELKObject P_Symbol_Plist (ELKObject sym) {
	Check_Type (sym, T_Symbol);
	return Copy_List (SYMBOL(sym)->plist);
}

int Hash (CHAR *str, int len) {
	register int h;
	register CHAR *p, *ep;

	h = 5 * len;
	if (len > 5)
		len = 5;
	for (p = str, ep = p+len; p < ep; ++p)
		h = (h << 2) ^ *p;
	return h & 017777777777;
}

void Define_Symbol (ELKObject *sym, const CHAR *name) {
	*sym = Intern (name);
	Func_Global_GC_Link (sym);
}

void Define_Variable (ELKObject *var, const CHAR *name, ELKObject init) {
	ELKObject frame, sym;
	GC_Node;

	GC_Link (init);
	sym = Intern (name);
	SYMBOL(sym)->value = init;
	frame = Add_Binding (Car (The_Environment), sym, init);
	*var = Car (frame);
	Car (The_Environment) = frame;
	Func_Global_GC_Link (var);
	GC_Unlink;
}

ELKObject Var_Get (ELKObject var) {
	return Cdr (var);
}

void Var_Set (ELKObject var, ELKObject val) {
	Cdr (var) = val;
	SYMBOL (Car (var))->value = val;
}

int Var_Is_True (ELKObject var) {
	var = Var_Get (var);
	return Truep (var);
}

unsigned long Symbols_To_Bits (ELKObject x, int mflag, SYMDESCR *stab) {
	register SYMDESCR *syms;
	register unsigned long mask = 0;
	ELKObject l, s;
	register CHAR *p;
	register int n;

	if (!mflag) Check_Type (x, T_Symbol);
	for (l = x; !Nullp (l); l = Cdr (l)) {
		if (mflag) {
			Check_Type (l, T_Pair);
			x = Car (l);
		}
		Check_Type (x, T_Symbol);
		s = SYMBOL(x)->name;
		p = STRING(s)->data;
		n = STRING(s)->size;
		for (syms = stab; syms->name; syms++)
			if (n && estrncmp (syms->name, p, n) == 0) break;
		if (syms->name == 0)
			Primitive_Error (x_("invalid argument: ~s"), x);
		mask |= syms->val;
		if (!mflag) break;
	}
	return mask;
}

ELKObject Bits_To_Symbols (unsigned long x, int mflag, SYMDESCR *stab) {
	register SYMDESCR *syms;
	ELKObject list, tail, cell;
	GC_Node2;

	if (mflag) {
		GC_Link2 (list, tail);
		for (list = tail = Null, syms = stab; syms->name; syms++)
			if ((x & syms->val) && syms->val != ~0) {
				ELKObject z;

				z = Intern (syms->name);
				cell = Cons (z, Null);
				if (Nullp (list))
					list = cell;
				else
					P_Set_Cdr (tail, cell);
				tail = cell;
			}
		GC_Unlink;
		return list;
	}
	for (syms = stab; syms->name; syms++)
		if (syms->val == x)
			return Intern (syms->name);
	return Null;
}
