/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */

#include "global.h"
#include "irsimnet.h"
#include "irsimglobals.h"
#include "irsimsubckt.h"

#ifdef USER_SUBCKT 

typedef struct _Sub
{
    CHAR	*name;			/* name of this sub-circuit */
    ufun	model;			/* function that models sub-circuit */
    ufun	init;			/* Function to call to init */
    int		inst;			/* Number of instantiations */
    struct _Sub	*next;			/* list of commands in bucket */
} SubCircuit;

CHAR pot2char[N_POTS] = { 'L' , 'x', 'X' , 'H' } ;

#define SUBSTBLSIZE	16

static SubCircuit *substbl[SUBSTBLSIZE];


userSubCircuit subs[] =
{
    {NULL,		NULL,		NULL}
};


static int subs_cmp(SubCircuit *a, SubCircuit *b)
{
    return(namesame(a->name, b->name));
}


void irsim_init_subs(void)
{
	register int n;
	register userSubCircuit *s;
	register SubCircuit *sl;

	for (n=0; n < SUBSTBLSIZE; n++)
		substbl[n] = NULL;
	for (s=subs; s->name != NULL; s += 1)
	{
		sl = (SubCircuit *) emalloc(sizeof(SubCircuit), sim_tool->cluster);
		sl->name = s->name;
		sl->model = s->model;
		sl->init = s->init;
		sl->inst = 0 ;
		n = HashSub(s->name);
		sl->next = substbl[n];
		substbl[n] = sl;
	}
}


static int HashSub(register CHAR  *name)
{
	register int  hashcode = 0;

	do
		hashcode = (hashcode << 1) ^ (*name | 0x20);
	while(*(++name));
	return(((hashcode >= 0) ? hashcode : ~hashcode) % SUBSTBLSIZE);
}

ufun irsim_subckt_instantiate(CHAR *sname, ufun *initfunP, int *instP)
{
	register int n;
	register SubCircuit *s;

	n = HashSub(sname);

	for (s = substbl[n]; s != NULL; s = s->next)
	{
		if (namesame(s->name, sname) == 0)
		{
			*initfunP = s->init;
			*instP = s->inst ;
			s->inst++;
			return(s->model);
		}
	}
	return((ufun)NULL);
}
	

#define INVERT(x)	((x == LOW) ? HIGH : ((x == HIGH) ? LOW : X))

void irsim_subckt_model_C(tptr t)
{
	int i;
	register short nins = t->subptr->nins;
	register short nouts = t->subptr->nouts;
	register nptr *nodes = t->subptr->nodes;
	uptr udata = t->subptr->udata;

	CHAR *in, *out;
	double *delay;

	in  = (CHAR *) emalloc(nins * SIZEOFCHAR, sim_tool->cluster);
	out = (CHAR *) emalloc(nouts * SIZEOFCHAR, sim_tool->cluster);
	delay = (double *) emalloc(nouts*sizeof(double), sim_tool->cluster);
	for (i = 0; i < nins; i++)
		in[i] = nodes[i]->npot;
	for (i = 0; i < nouts; i++)
		out[i] = nodes[i+nins]->npot;
	(*t->subptr->model)(in, out, delay, udata);
	for (i = 0; i < nouts; i++)
	{
		if (out[i] != X_X)
		{
			irsim_QueueFVal(nodes[i+nins],INVERT(out[i]),
				(double) 1.0, delay[i]);
			irsim_QueueFVal(nodes[i+nins+nouts],INVERT(out[i]), 
				(double) 1.0, delay[i]);
		} else
		{
			irsim_QueueFVal(nodes[i+nins], HIGH,
				(double) 1.0, delay[i]);
			irsim_QueueFVal(nodes[i+nins+nouts], LOW,
				(double) 1.0, delay[i]);
		}
	}
	free(in); 
	free(out); 
	free(delay);
}
#endif
